/* scratchpad - Clipboard-based prototype of the Netscape Scratchpad system
 * Copyright (C) 1998-2000 by Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "dnd.h"
#include "mozilla_remote.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkbox.h>
#include <gtk/gtkclist.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkhseparator.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkselection.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkwindow.h>
#include <glib.h>

/* where to find netscape; add a fully qualified path if Netscape isn't in
 * your $PATH.
 */
#define NETSCAPE "netscape"

static GtkCList *dogears, *pending;
static GdkAtom ctext_atom = GDK_NONE;
char *url_prefix[] = { "http://", "ftp://", "file:/", "gopher://", 0 };

char *progname;
char *expected_mozilla_version = "4.61";

void netscape_open(char *s);
int netscape_remote(char *s);

/* Returns the currently selected row (counting from 0) in the given list. */
int list_get_selected(GtkCList *list) {
  int ret;

  g_return_val_if_fail(list != NULL, -1);
  g_return_val_if_fail(GTK_IS_CLIST(list), -1);
  g_return_val_if_fail(list->selection != NULL, -1);

  ret = GPOINTER_TO_INT(list->selection->data);
  if (ret > list->rows) ret = -1;
  return ret;
}

/* Returns the text associated with the given row (n) in the given list. */
char *list_get_text(GtkCList *list, int n) {
  char *s;

  g_return_val_if_fail(list != NULL, NULL);
  g_return_val_if_fail(GTK_IS_CLIST(list), NULL);
  g_return_val_if_fail(n >= 0 && n < list->rows, NULL);

  gtk_clist_get_text(list, n, 0, &s);
  return s;
}

/* Remove the given entry (n) from the given list (list). */
void remove_entry(GtkCList *list, int n) {
  g_return_if_fail(list != NULL);
  g_return_if_fail(GTK_IS_CLIST(list));
  g_return_if_fail(n >= 0 && n < list->rows);
  gtk_clist_remove(list, n);
  if (list->rows <= n)
    gtk_clist_select_row(list, list->rows-1, 0);
  else
    gtk_clist_select_row(list, n, 0);
}

/* Remove the currently selected entry from the given list (list). */
void remove_current(GtkCList *list) {
  remove_entry(list, list_get_selected(list));
}

/* Display the given entry (n) from the given list (list).  Delete the
 * entry if the given list is the pending window.
 */
void do_pop(GtkCList *list, int n) {
  g_return_if_fail(list != NULL);
  g_return_if_fail(GTK_IS_CLIST(list));
  if (n >= list->rows) return;
  netscape_open(list_get_text(list, n));
  if (list == pending)
    gtk_clist_remove(list, n);
}

/* Display the current entry from the given list (list).  Delete the
 * entry if the given list is the pending window.
 */
void pop_current(GtkCList *list) {
  do_pop(list, list_get_selected(list));
}

/* (callback) Display the and remote current entry from the pending list. */
void pop(void) {
  do_pop(pending, 0);
}

/* (callback) Add a new dogear for the current URL. */
void new_dogear(void) {
  char *url = mozilla_remote_get_current_url(gdk_display);
  g_return_if_fail(url != NULL);
  gtk_clist_append(dogears, &url);
  free(url);
}

/* Add all URLs found in the given string (s) to the pending list.
 * Returns the number of URLs added.
 */
int parse_and_add(char *s) {
  int i, ret = 0;
  char *t, *minp, *p = s;

  g_return_val_if_fail(s != NULL, 0);

  while (1) {
    minp = NULL;
    for (i=0; url_prefix[i]; i++) {
      t = strstr(p, url_prefix[i]);
      if (t && (!minp || t < minp)) minp = t;
    }
    if (!minp) return ret;
    t = strpbrk(minp, " \t\n\r\v\f\"\'>");
    if (t) { *t = 0; }
    gtk_clist_append(pending, &minp);
    ret++;
    if (!t) return ret;
    p = t+1;
  }
}

/* (callback) Some other application is advertising a selection; handle it. */
void handle_selection(GtkWidget *widget, GtkSelectionData *selection_data) {
  enum {INVALID, STRING, CTEXT} type;

  g_return_if_fail(widget != NULL);
  g_return_if_fail(widget == GTK_WIDGET(pending));
  g_return_if_fail(selection_data != NULL);

  if (selection_data->type == GDK_TARGET_STRING)
    type = STRING;
  else if (selection_data->type == ctext_atom)
    type = CTEXT;
  else
    type = INVALID;

  if (type == INVALID || selection_data->length < 0) {
    if (selection_data->target != GDK_TARGET_STRING)
      gtk_selection_convert(widget, selection_data->selection,
        GDK_TARGET_STRING, GDK_CURRENT_TIME);
    /* avoid infinite loop: handle_selection will be called with the new,
     * converted selection.
     */
    return;
  }

  switch (type) {
    case STRING:
      selection_data->data[selection_data->length] = 0;
      if (parse_and_add(selection_data->data))
        gtk_selection_owner_set(NULL, GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
      break;
    case CTEXT:
      {
        gchar **list;
        gint count;
        gint i;

        count = gdk_text_property_to_text_list(selection_data->type,
          selection_data->format, selection_data->data,
          selection_data->length, &list);
        for (i=0; i<count; i++) {
          if (parse_and_add(list[i]))
            gtk_selection_owner_set(NULL, GDK_SELECTION_PRIMARY,
              GDK_CURRENT_TIME);
        }
        if (count > 0)
          gdk_free_text_list(list);
      }
      break;
    case INVALID:    /* suppress compiler warning */
      break;
  }
}

/* (timer callback) Check to see if any new selections are being exported. */
int selection_poll(gpointer data) {
  gtk_selection_convert(GTK_WIDGET(pending), GDK_SELECTION_PRIMARY,
    ctext_atom, time(0));
  return TRUE;
}

/* (callback) A row has been selected; if this is a double-click, view
 * (and possibly remove) the URL.
 */
void row_selected(GtkWidget *w, int row, int col, GdkEventButton *ev,
    gpointer d) {
  g_return_if_fail(w != NULL);
  g_return_if_fail(GTK_IS_CLIST(w));
  if (ev)
    switch (ev->type) {
      case GDK_2BUTTON_PRESS:
        do_pop(GTK_CLIST(w), row);
        break;
      default: ;
    }
}

/* (callback) Handle keyboard events for both lists. */
void key_pressed(GtkWidget *w, GdkEventKey *ev) {
  g_return_if_fail(w != NULL);
  g_return_if_fail(GTK_IS_CLIST(w));
  g_return_if_fail(ev != NULL);

  switch (ev->keyval) {
    case GDK_Return:
    case GDK_KP_Enter:
      pop_current(GTK_CLIST(w));
      break;
    case GDK_d:
      if (!(ev->state & GDK_CONTROL_MASK)) break;
      /* otherwise (if CTRL+D) carry over into GDK_Delete */
    case GDK_Delete:
    case GDK_KP_Delete:
      if (GTK_CLIST(w)->rows > 0)
        remove_current(GTK_CLIST(w));
      break;
  }
}

static GtkTooltips *tooltips;
/* Add a button with the given label (label) and tooltip (tip, which may
 * be NULL) to the given hbox or vbox (box) with the given callback
 * (function) and callback argument (arg).
 */
void add_button(char *label, char *tip, GtkWidget *box, void *function,
    gpointer arg) {
  GtkWidget *button;

  g_return_if_fail(label != NULL);
  g_return_if_fail(box != NULL);
  g_return_if_fail(GTK_IS_BOX(box));
  g_return_if_fail(function != NULL);

  button = gtk_button_new_with_label(label);
  if (tip) gtk_tooltips_set_tip(tooltips, button, tip, NULL);
  gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
    GTK_SIGNAL_FUNC(function), arg);
}

/* Create and return a GtkCList and insert it, along with the given label
 * (label) into the given vbox (box).
 */
GtkCList *make_clist(char *label, GtkWidget *box) {
  GtkCList *ret;
  GtkWidget *swin;

  g_return_val_if_fail(label != NULL, NULL);
  g_return_val_if_fail(box != NULL, NULL);
  g_return_val_if_fail(GTK_IS_VBOX(box), NULL);

  gtk_box_pack_start(GTK_BOX(box), gtk_label_new(label), FALSE, FALSE, 0);

  ret = GTK_CLIST(gtk_clist_new(1));
  gtk_clist_set_column_auto_resize(ret, 0, TRUE);

  gtk_signal_connect(GTK_OBJECT(ret), "select_row",
    GTK_SIGNAL_FUNC(row_selected), 0);
  gtk_signal_connect(GTK_OBJECT(ret), "key_press_event",
    GTK_SIGNAL_FUNC(key_pressed), 0);

  swin = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(swin), GTK_WIDGET(ret));
  gtk_box_pack_start(GTK_BOX(box), swin, TRUE, TRUE, 0);

  return ret;
}

/* Create and return the main Scratchpad window with the given width
 * and height.
 */
GtkWidget *make_main_window(int width, int height) {
  GtkWidget *vbox, *hbox, *win;

  g_return_val_if_fail(width > 0 && width < 5000, NULL);
  g_return_val_if_fail(height > 0 && height < 5000, NULL);
  
  win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_policy(GTK_WINDOW(win), TRUE, TRUE, TRUE);
  gtk_window_set_title(GTK_WINDOW(win), "Scratchpad");
  gtk_container_set_border_width(GTK_CONTAINER(win), 5);
  gtk_widget_set_usize(win, width, height);
  gtk_signal_connect(GTK_OBJECT(win), "destroy", gtk_main_quit, 0);

  tooltips = gtk_tooltips_new();
  gtk_object_set_data(GTK_OBJECT(win), "tooltips", tooltips);

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(win), vbox);

  hbox = gtk_hbox_new(TRUE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  add_button("Back",      "Go to the previous page",
    hbox, netscape_remote, (gpointer)"back");
  add_button("Forward",   "Go to the next page",
    hbox, netscape_remote, (gpointer)"forward");
  add_button("Stop",      "Stop the current transfer",
    hbox, netscape_remote, (gpointer)"stopLoading");

  gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 5);

  dogears = make_clist("Dogears", vbox);
  init_dnd(GTK_WIDGET(dogears));

  gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 5);

  pending = make_clist("Pending links", vbox);
  gtk_signal_connect(GTK_OBJECT(pending), "selection_received",
    GTK_SIGNAL_FUNC(handle_selection), 0);
  init_dnd(GTK_WIDGET(pending));

  gtk_box_pack_start(GTK_BOX(vbox), gtk_hseparator_new(), FALSE, FALSE, 5);

  hbox = gtk_hbox_new(TRUE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  add_button("Next pending link", "Load and remove the next pendig link",
    hbox, pop,           0);
  add_button("Add dogear",        "Temporarily bookmark the current page",
    hbox, new_dogear,    0);
  add_button("Quit",              "Close Scratchpad",
    hbox, gtk_main_quit, 0);

  return win;
}

int main(int argc, char **argv) {
  int timer;
  GtkWidget *win;

  progname = argv[0];
  gtk_init(&argc, &argv);

  /* very clumsy: check for new selections every 0.1 seconds.  There does
   * not seem to be a better way to find out when some other application
   * has exported a new selection.  xev (the X event viewer) does not show
   * any events received when another app exports a selection, so odds are
   * there's no way to get notified other than to poll.  Ugly.
   * This could all be avoided if we could hack Netscape to send us some
   * sort of event.
   */
  timer = gtk_timeout_add(100, selection_poll, 0);
  if (ctext_atom == GDK_NONE)
    ctext_atom = gdk_atom_intern("COMPOUND_TEXT", FALSE);

  win = make_main_window(350, 450);
  gtk_widget_show_all(win);

  gtk_main();

  gtk_timeout_remove(timer);

  return 0;
}

/* Open the given URL (s) in a Netscape window.  A direct connection is
 * tried first; if that fails, we exec "netscape -remote ..."
 */
void netscape_open(char *s) {
  char *buf;
  g_return_if_fail(s != NULL);
  buf = malloc(strlen(s) + 8 + 1 + 1);
  sprintf(buf, "openURL(%s)", s);
  if (netscape_remote(buf) == -1) {
    switch (fork()) {
      case -1:
        perror("fork");
        free(buf);
        return;
        break;
      case 0:
        /* prevents us getting events intended for Netscape */
        close(ConnectionNumber(gdk_display));
        execlp(NETSCAPE, NETSCAPE, s, 0);
        perror(NETSCAPE);
        exit(-1);
        break;
      default:
        break;
    }
  }
  free(buf);
}

/* Execute the given mozilla-remote command (s). */
int netscape_remote(char *s) {
  char *c1[3] = { "-noraise", s, 0 };
  char **c = c1;
  g_return_val_if_fail(s != NULL, -1);
  if (mozilla_remote_commands(gdk_display, 0, c) == -1)
    return -1;
  return 0;
}
